#include <SPI.h>
#include "PCF8574.h"  // https://github.com/xreef/PCF8574_library
#include <si5351.h>
#include <JTEncode.h>
#include <TimeLib.h>
#include <NTPtimeESP.h>
#include <ESP8266WiFi.h>
#include <WiFiClient.h>
#include <U8g2lib.h>   // https://github.com/olikraus/u8g2

// Set i2c address PCF8574
PCF8574 pcf8574(0x20, 2, 14);
U8G2_UC1609_SLG19264_F_4W_SW_SPI u8g2(U8G2_R0, /* clock=*/ 4, /* data=*/ 0, /* cs=*/ 5, /* dc=*/ 12, /* reset=*/ 13);

#define TONE_SPACING            146           // ~1.46 Hz
#define WSPR_DELAY              683          // Delay value for WSPR
#define WSPR_CTC                10672         // CTC value for WSPR
#define SYMBOL_COUNT            WSPR_SYMBOL_COUNT
#define CORRECTION              7526             // Change this for your ref osc

//****************************************************
//* SYNCRONIZE SYSTEM TIME with NTP SERVERS
//* need to be modified, obsolete library in use...
//****************************************************
#define SEND_INTV     10
#define RECV_TIMEOUT  10

//****************************************************
//list of frequencies for jack 1 (clock0 of si5351) output
#define MAXCH      7 //Change this according to the number of bands inserted below
unsigned long freq0[] = {28126100UL, 24926100UL, 21096100UL, 18106100UL, 14097100UL, 10140200UL, 7040100UL}; //CHANGE THIS: is the freq of multiband output on jack 1 (clock0)
uint8_t relay0[] =      {    3,         2,          2,           1,          1,          0,         0};

#define SI5351_REF     26000000UL  //change this to the frequency of the crystal on your si5351’s PCB, usually 25 or 27 MHz

char call[7] = "RA4FHE";                        // Change this
char loc[5] = "LO23";                           // Change this
uint8_t dbm = 23; // Min = 0 dBm, Max = 43 dBm, steps 0,3,7,10,13,17,20,23,27,30,33,37,40,43
uint8_t tx_buffer[SYMBOL_COUNT];

const char* ssid = "Infotex";       //SSID of your Wifi network: Change this
const char* password = "12345678";      //Wi-Fi Password: Change this

//**** Sync the soft clock every 12 hours
#define NTPSYNC_DELAY  6

//**** NTP Server to use
const char* NTP_Server = "ntp4.ntp-servers.net";

//**** Your time zone UTC related (floating point number)
#define TIME_ZONE 3.0f

Si5351 si5351;
//some system variable. Anything to touch from here
// Global variables
JTEncode jtencode;
int ch = 0; // automatic frequency switch index
bool warmup = 0;

NTPtime NTPch(NTP_Server);
strDateTime dateTime;

// --------------------------------------
// epochUnixNTP set the UNIX time
// number of seconds sice Jan 1 1970
// --------------------------------------

time_t epochUnixNTP()
{
  Serial.println(">>>>>>>> Time Sync function called <<<<<<<<<");

  //**** BIG ISSUE: in case of poor connection, we risk to remain in this loop forever
  NTPch.setSendInterval(SEND_INTV);
  NTPch.setRecvTimeout(RECV_TIMEOUT);
  do
  {
    dateTime = NTPch.getNTPtime(TIME_ZONE, 1);
    delay(1);
  }
  while (!dateTime.valid);
  NTPch.printDateTime(dateTime);
  setTime(dateTime.hour, dateTime.minute, dateTime.second, dateTime.day, dateTime.month, dateTime.year);
  Serial.println(now());
  return 0;
}

void printTx(bool tx)
{
  u8g2.setFont(u8g2_font_open_iconic_all_2x_t);
  u8g2.setFontMode(0);
  u8g2.setDrawColor(0);
  u8g2.drawBox(170, 0, 20, 20);
  u8g2.setDrawColor(1);
  u8g2.setCursor(190, 28);
  if (tx) {
    u8g2.drawGlyph(172, 18, 233);
  } else
  {
    u8g2.drawGlyph(172, 18, 121);
  }
}

void printFreq()
{
  u8g2.setFont(u8g2_font_crox5tb_tr);
  u8g2.setFontMode(0);
  u8g2.setDrawColor(0);
  u8g2.drawBox(0, 43, 192, 20);
  u8g2.setDrawColor(1);
  u8g2.setCursor(26, 62);
  u8g2.print(freq0[ch]);
  u8g2.println(" Hz");
  u8g2.sendBuffer();
}

void printTime(bool printSync)
{
  u8g2.setFont(u8g2_font_crox5tb_tr);
  u8g2.setFontMode(0);
  u8g2.setDrawColor(0);
  u8g2.drawBox(45, 0, 111, 20);
  u8g2.setDrawColor(1);
  u8g2.setCursor(46, 18);
  if (printSync) {
    u8g2.print("Get Time");
  } else
  {
    u8g2.printf("%02d:%02d:%02d", hour(), minute(), second());
  }
  u8g2.sendBuffer();
}

void setFilter(uint8_t relay)
{
  for (uint8_t i = 0; i < 4; ++i) {
    if (i != relay) {
      pcf8574.digitalWrite(i, LOW);
    }
  }
  pcf8574.digitalWrite(relay, HIGH);
}

// Loop through the string, transmitting one character at a time.
void encode()
{
  uint8_t i;

  jtencode.wspr_encode(call, loc, dbm, tx_buffer);

  si5351.set_clock_pwr(SI5351_CLK0, 1);

  Serial.println("TX ON");
  printTx(true);
  printFreq();

  // Now do the rest of the message
  uint32_t period = millis() - WSPR_DELAY;
  uint8_t sec = second();
  i = 0;
  while (i < SYMBOL_COUNT) {
    uint32_t curMill = millis();
    if (curMill - period >= WSPR_DELAY) {
      period = curMill;
      setFilter(relay0[ch]);
      si5351.set_freq((freq0[ch] * 100) + (tx_buffer[i] * TONE_SPACING), SI5351_CLK0);
      ++i;
    }
    if (sec != second()) {
      sec = second();
      printTime(false);
    }
  }

  // Turn off the output
  si5351.set_clock_pwr(SI5351_CLK0, 0);

  Serial.println("TX OFF");
  printTx(false);

  //change band on clock0 for next tx
  ch++;
  if (ch == MAXCH) ch = 0; //reset band index to start again from first band the in array
  printFreq();
}

void printWiFi()
{
  u8g2.setFont(u8g2_font_open_iconic_all_2x_t);
  u8g2.setFontMode(0);
  u8g2.setDrawColor(0);
  u8g2.drawBox(0, 0, 20, 20);
  u8g2.setDrawColor(1);
  u8g2.setCursor(1, 28);
  if (WiFi.status() == WL_CONNECTED) {
    u8g2.drawGlyph(2, 18, 247);
  } else
  {
    u8g2.drawGlyph(2, 18, 121);
  }
}

void ssidConnect()
{
  Serial.println(ssid);
  Serial.println(password);
  WiFi.begin(ssid, password);

  while (WiFi.status() != WL_CONNECTED) {
    delay(700);
    Serial.print(".");
  }

  Serial.println();
  Serial.print(F("Connected to "));
  Serial.println(ssid);
  printWiFi();
  Serial.print(F("IP address: "));
  Serial.println(WiFi.localIP());
}

void setup()
{
  Serial.begin(74880); while (!Serial);
  Serial.println("COM setup successful");
  delay(10);
  u8g2.begin();
  u8g2.enableUTF8Print();
  u8g2.setFont(u8g2_font_crox5tb_tr);
  u8g2.setFontDirection(0);
  u8g2.clearBuffer();
  u8g2.setCursor(8, 36);
  u8g2.print("Connect to Wi-Fi");
  u8g2.sendBuffer();
  delay(1000);
  u8g2.clearBuffer();

  WiFi.mode(WIFI_STA);
  // connect to WiFi network
  ssidConnect();
  printTime(true);

  // Set time sync provider
  setSyncProvider(epochUnixNTP);  //set function to call when sync required
  setSyncInterval(3600 * 24); // Sync time every 24 hour

  pcf8574.begin();
  // Set pinMode to OUTPUT
  pcf8574.pinMode(0, OUTPUT, LOW);
  pcf8574.pinMode(1, OUTPUT, LOW);
  pcf8574.pinMode(2, OUTPUT, LOW);
  pcf8574.pinMode(3, OUTPUT, LOW);

  // Initialize the Si5351
  // Change the 2nd parameter in init if using a ref osc other
  // than 25 MHz
  Serial.println("start radio module setup");
  si5351.init(SI5351_CRYSTAL_LOAD_0PF, SI5351_REF, CORRECTION);
  Serial.println("Module intializated");

  // Set CLK0 out  put
  setFilter(relay0[ch]);
  si5351.set_freq(freq0[ch] * 100, SI5351_CLK0);
  si5351.drive_strength(SI5351_CLK0, SI5351_DRIVE_4MA); // Set for max power
  si5351.set_clock_pwr(SI5351_CLK0, 0); // Disable the clock initially
  Serial.println("Radio Module setup successful");
  Serial.println("Entering loop...");
}

void loop()
{
  printTime(false);
  printWiFi();

  if (minute() % 2 == 0 && second() == 0)
  {
    encode();
  }
  delay(50);
}
